//shading:         phong
//lighting model:  blinn
//lighting type:   directional

// --------------------------------------------------------------------------------------------------
// PARAMETERS:
// --------------------------------------------------------------------------------------------------

//transforms
float4x4 tW: WORLD;        //the models world matrix
float4x4 tV: VIEW;         //view matrix as set via Renderer (EX9)
float4x4 tWV: WORLDVIEW;
float4x4 tWVP: WORLDVIEWPROJECTION;
float4x4 tP: PROJECTION;   //projection matrix as set via Renderer (EX9)

//light properties
float3 lDir <string uiname="Light Direction";> = {0, -5, 2};        //light direction in world space
float4 lAmb  : COLOR <String uiname="Ambient Color";>  = {0.15, 0.15, 0.15, 1};
float4 lDiff : COLOR <String uiname="Diffuse Color";>  = {0.85, 0.85, 0.85, 1};
float4 lSpec : COLOR <String uiname="Specular Color";> = {0.35, 0.35, 0.35, 1};
float lPower <String uiname="Power"; float uimin=0.0;> = 25.0;     //shininess of specular highlight

//texture
texture Tex <string uiname="Texture";>;
sampler Samp = sampler_state    //sampler for doing the texture-lookup
{
    Texture   = (Tex);          //apply a texture to the sampler
    MipFilter = LINEAR;         //sampler states
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

float4x4 tTex: TEXTUREMATRIX <string uiname="Texture Transform";>;
float4x4 tColor <string uiname="Color Transform";>;

struct vs2ps
{
    float4 PosWVP: POSITION;
    float4 TexCd : TEXCOORD0;
    float3 LightDirV: TEXCOORD1;
    float2 UV: TEXCOORD2;
    float3 ViewDirV: TEXCOORD3;
};

// --------------------------------------------------------------------------------------------------
// VERTEXSHADERS
// --------------------------------------------------------------------------------------------------

float gridSpaceX;
float gridSpaceY;

float gridOffsetX;
float gridOffsetY;

float gridScaleX = 1;
float gridScaleY = 1;

float3 lemniscape(float u, float v)
{
    // \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
    // http://local.wasp.uwa.edu.au/~pbourke/surfaces_curves/lemniscape/
    /*
    x = cos(v) sqrt(|sin(2 u)|) cos(u)
    y = cos(v) sqrt(|sin(2 u)|) sin(u)
    z = x2 - y2 + 2 x y tan2(v)
    
    0 <= u <= pi, 0 <= v <= pi
    */
    
    float x = cos(v)*sqrt(abs(sin(2*u)))*cos(u);
    float y = cos(v)*sqrt(abs(sin(2*u)))*sin(u);
    float z = x*x - y*y + 2*x*y*pow(tan(v),2);

    return float3(x, y, z);
}

vs2ps VS(
    float4 PosO: POSITION,
    float4 TexCd : TEXCOORD0)
{
    //inititalize all fields of output struct with 0
    vs2ps Out = (vs2ps)0;
    
    float x, y, z, u, v, pi;
    pi = 3.14159265;
    
    //map u and v
    gridScaleX *= pi;
    gridScaleY *= pi;
    
    u = (PosO.x + gridOffsetX) * gridScaleX;
    v = (PosO.y + gridOffsetY) * gridScaleY;
    
    Out.UV  = float2(u, v);
    
    //get position
    PosO.xyz = lemniscape(u, v);
    PosO.w = 1;
    
    // ////////////////////////////////////////////////////////////////////

    // \\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
    // http://local.wasp.uwa.edu.au/~pbourke/surfaces_curves/monkey/
    //PosO.z = pow(PosO.x,3)-3*PosO.x*pow(PosO.y,2);
    // ////////////////////////////////////////////////////////////////////

    //inverse light direction in view space
    Out.LightDirV = normalize(-mul(lDir, tV));
    
    //position (projected)
    Out.PosWVP  = mul(PosO, tWVP);
    Out.TexCd = mul(TexCd, tTex);
    Out.ViewDirV = -normalize(mul(PosO, tWV));
    return Out;
}

// --------------------------------------------------------------------------------------------------
// PIXELSHADERS:
// --------------------------------------------------------------------------------------------------

float4 PS(vs2ps In): COLOR
{
    float u1, v1, u2, v2;
    float3 tang, bitang;
    float3 NormO, NormV;
    
    //to get neighbour in u direction
    u1 = In.UV.x + gridSpaceX*gridScaleX; //(In.PosO.x + gridSpaceX + gridOffsetX) * gridScaleX;
    u2 = In.UV.x - gridSpaceX*gridScaleX; //(In.PosO.x - gridSpaceX + gridOffsetX) * gridScaleX;

    //to get neighbour in v direction
    v1 = In.UV.y + gridSpaceY*gridScaleY; //(In.PosO.y + gridSpaceY + gridOffsetY) * gridScaleY;
    v2 = In.UV.y - gridSpaceY*gridScaleY; //(In.PosO.y - gridSpaceY + gridOffsetY) * gridScaleY;

    //get tangents
    tang   = (lemniscape(u1, In.UV.y) - lemniscape(u2, In.UV.y));
    bitang = (lemniscape(In.UV.x, v1) - lemniscape(In.UV.x, v2));

    //get normal
    NormO = (cross(tang, bitang));

    //normal in view space
    NormV = mul(float4(NormO, 0), tWV);
    NormV = normalize(NormV);

    //In.TexCd = In.TexCd / In.TexCd.w; // for perpective texture projections (e.g. shadow maps) ps_2_0

    lAmb.a = 1;
    //halfvector
    float3 H = normalize(In.ViewDirV + In.LightDirV);

    //compute blinn lighting
    float3 shades = lit(dot(NormV, In.LightDirV), dot(NormV, H), lPower);

    float4 diff = lDiff * shades.y;
    diff.a = 1;

    //reflection vector (view space)
    float3 R = normalize(2 * dot(NormV, In.LightDirV) * NormV - In.LightDirV);
    //normalized view direction (view space)
    float3 V = normalize(In.ViewDirV);

    //calculate specular light
    float4 spec = pow(max(dot(R, V),0), lPower*.2) * lSpec;

    float4 col = tex2D(Samp, In.TexCd);
    col.rgb *= (lAmb + diff) + spec;
    return mul(col, tColor);
}


// --------------------------------------------------------------------------------------------------
// TECHNIQUES:
// --------------------------------------------------------------------------------------------------

technique TPhongDirectional
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_1 VS();
        PixelShader = compile ps_2_a PS();
    }
}

technique TFallbackGouraudDirectionalFF
{
    pass P0
    {
        //transformations
        WorldTransform[0]   = (tW);
        ViewTransform       = (tV);
        ProjectionTransform = (tP);

        //material
        MaterialAmbient  = {1, 1, 1, 1};
        MaterialDiffuse  = {1, 1, 1, 1};
        MaterialSpecular = {1, 1, 1, 1};
        MaterialPower    = (lPower);

        //texturing
        Sampler[0] = (Samp);
        TextureTransform[0] = (tTex);
        TexCoordIndex[0] = 0;
        TextureTransformFlags[0] = COUNT2;
        //Wrap0 = U;  // useful when mesh is round like a sphere

        //lighting
        LightEnable[0] = TRUE;
        Lighting       = TRUE;
        SpecularEnable = TRUE;

        LightType[0]     = DIRECTIONAL;
        LightAmbient[0]  = (lAmb);
        LightDiffuse[0]  = (lDiff);
        LightSpecular[0] = (lSpec);
        LightDirection[0] = (lDir);

        //shading
        ShadeMode = GOURAUD;
        VertexShader = NULL;
        PixelShader  = NULL;
    }
}
